/*****************************************************************************
*                              18fSensors.c                                  *
******************************************************************************

Implements communication between the Controller board and the Sensor boards
via I2C protocol.

Date:		Author:				Comment:
------------------------------------------------------------------------------
2006-2-6	Jonathan Karpick	Assembly version
2006-5-8	Jonathan Karpick	Port to C
2006-7-5	Barrett Heyneman	General Code cleanup, moved sensor specific
								functionality to this file.
2006-8-4	Barrett Heyneman	Changed the scope of the sensor reading var.
								to private and implemented (partial) batch
								accessors.  Also, added the (snub) function
								for turning raw readings into signed int force
								values.
2006-8-7	Barrett Heyneman	Added a separate SNS_ResetI2C function, added
								initialization functionality for sensor
								calibration to InitSensors, and filled in snub
								Convert2Force function.
2006-8-9	Barrett Heyneman	Created new types for holding sensor data, and
								reworked old/added new accessors/variables to
								use the	new storage types.
2006-8-30	Salomon Trujillo	Added a low-pass filter to the sensor data.
*****************************************************************************/

#include "18fSensors.h"
#include "18f4520_flash.h"
#include "FileReader.h"
#include "ControllerMath.h"
#include "PhysicalConfig.h"
#include <p18cxxx.h>
#include <i2c.h>


/*****************************************************************************
					PRIVATE TYPES, MACROS, DEFINITIONS
*****************************************************************************/
/* Addresses for the sensors.  Note: addresses can only be 0-127.  Also, the
   following addresses are reserved by the I2C protocol:
   00000xxx (0-7)
   01111xxx (120-127)
   */
#define FL_STROKE_ADDR		0b00001000
#define FL_WING_ADDR		0b00001001
#define FR_STROKE_ADDR		0b00001010
#define FR_WING_ADDR		0b00001011
#define BL_STROKE_ADDR		0b00001100
#define BL_WING_ADDR		0b00001101
#define BR_STROKE_ADDR		0b00001110
#define BR_WING_ADDR		0b00001111

/* How many times to loop while waiting for SSPIF to be set.  Allows us to
 * quit waiting for a sensor if something screws up the I2C communication. */
#define TIME_OUT_CNT	100		//2000 last time we tested

#define MASTER_MODE		0x08	// Value in SSPCON1 to set to master mode

/* Default values for the static arrays */
#define RAW_VALUE		0x80		// unsigned char
#define SENSOR_ZERO		0x80		// unsigned char
#define SENSOR_SLOPE	0x0100		// Number_t

/*****************************************************************************
							PRIVATE VARIABLES
*****************************************************************************/

/* Accesses sensor addresses by sensor */
static SensorRaw_t SensorAddr;

/* Raw sensor readings */
static SensorRaw_t SensorReadings;

/* Converted to force values */
static SensorNumber_t ForceValues;

/* Zero points for all the sensors */
static SensorRaw_t SensorZeros;

/* Slope for converting raw sensor data to force values */
static SensorNumber_t SensorSlopes;

/* History for low-pass filter */
static signed int filter[NUM_SENSOR_TYPES][NUM_LEGS];

/*****************************************************************************
						PRIVATE FUNCTION PROTOTYPES
*****************************************************************************/

/* Converts raw readings to force values */
static void Convert2Force(void);

/* Add a low-pass filter to the sensor data */
static void LowPassFilter(void);


/*****************************************************************************
							FUNCTION DEFINITIONS	
*****************************************************************************/

/****************************** InitSensors **********************************
Initializes communication with the sensors by setting up the I2C communication
on the MSSP hardware.  Also reads in sensor calibration data from the gait
file; if there isn't enough data for the number of sensors specified here,
then all uninitialized sensor calibration data defaults to sensor zero and
sensor slope values defined above (128 and 1, respectively).

Last Modified:
Barrett Heyneman	2006-8-7

*****************************************************************************/

void SNS_InitSensors(void)
{	
	unsigned char i;
	unsigned char j;
	unsigned char n;
	near unsigned short long data_loc;
	
	/* Initialize the zeros, slopes, and raw readings, in case there aren't
	 * enough in the gait file or one of the sensors doesn't respond.
	 */
	for (i = 0; i < NUM_SENSOR_TYPES; i++)
	{
		for (j = 0; j < NUM_LEGS; j++)
		{
			SensorReadings.asTable[i][j] = RAW_VALUE;
			SensorZeros.asTable[i][j] = SENSOR_ZERO;
			SensorSlopes.asTable[i][j].val = SENSOR_SLOPE;
			filter[i][j] = 0;
		}
	}
		
	/* Initialize the sensor addresses */
	SensorAddr.byName.FL_Stroke = FL_STROKE_ADDR;
	SensorAddr.byName.FR_Stroke = FR_STROKE_ADDR;
	SensorAddr.byName.BL_Stroke = BL_STROKE_ADDR;
	SensorAddr.byName.BR_Stroke = BR_STROKE_ADDR;
	
	/* See if there is sensor calibration data */
	data_loc = (near unsigned short long)FR_FindEntry(SENSOR_CLBRTN_TABLE);
	
	/* If there is, copy it into RAM */
	if (data_loc != NO_ENTRY_FOUND)
	{
		n = ReadFlashByte(data_loc++);
		
		/* We only want to read in data for min(n, NUM_SENSORS) */
		if (n > (NUM_SENSOR_TYPES * NUM_LEGS))
			n = (NUM_SENSOR_TYPES * NUM_LEGS);
		
		ReadFlash(data_loc, &SensorZeros, n);
		data_loc += n;
		
		ReadFlash(data_loc, &SensorSlopes, n*NUMBER_SIZE);
	}
		
	
	/* Set the I2C baud rate generator value */
	SSPADD = I2C_NUMBR;
	
	/* Set the ports to outputs (this is the workaround in case the pins get
	 * latched high) */
	TRISCbits.TRISC3 = 0;
	TRISCbits.TRISC4 = 0;

	/* Ensure the bits are correctly set for I2C operation. */
	PORTCbits.SCL = 0;
	PORTCbits.SDA = 0;
	
	/* Now set to inputs. */
	TRISCbits.TRISC3 = 1;
	TRISCbits.TRISC4 = 1;
	
	/* Clear and reset all the relevant registers. */
	SNS_ResetI2C();
	
	/* Poll the sensors once to get good data. */
	SNS_PollSensors();
}


/******************************* SNS_ResetI2C ************************************
Resets the I2C module by turning it off, then resetting the values in all the
registers.  Does not implement the latch work around or value initialization
which is present in InitSensors.

Last Modified:
Barrett Heyneman	2006-8-7

*****************************************************************************/

void SNS_ResetI2C(void)
{
	///TEST
	PORTCbits.RC0 ^= 1;
	
	/* Clear all the relevant control bits. */
	SSPSTAT &= 0x3F;
	SSPCON1 = 0x00;
	SSPCON2 = 0x00;
	
	/* Set to maseter mode, regular speed, and enable */
	SSPCON1 |= MASTER_MODE;
	SSPSTATbits.SMP = 1;	
	SSPCON1bits.SSPEN = 1;
}


/****************************** PollSensors **********************************
Polls all the sensors and stores the values in the public SensorReading
variable.  This function is blocking (since the control loop is interrupt
driven and we want the sensors to update as often as possible) instead of
being interrupt driven.  This function should *NOT* be called within any
interrupt

At last check the best case (all 4 sensors respond) execution time is 246 us,
and the worst case (all sensors requested cause the master to hang in the last
loop) execution time is 1.044 ms.

Last Modified:
Barrett Heyneman	2006-8-8

*****************************************************************************/

void SNS_PollSensors(void)
{
	unsigned char type;
	unsigned char leg;
	unsigned int tries = 0;
	
	
	for (type = 0; type < NUM_SENSOR_TYPES; type++)
	{
		for (leg = 0; leg < NUM_LEGS; leg++)
		{			
			/* Try to generate a start.  If it hangs too long in the loop, try the
			 * next sensor.
			 */
			PIR1bits.SSPIF = 0;
			tries = 0;
			SSPCON2bits.SEN = 1;
			while (!PIR1bits.SSPIF)
			{
				if(++tries > TIME_OUT_CNT)
					break;
			}
			if (tries > TIME_OUT_CNT)
			{
				SNS_ResetI2C();
				continue;
			}
			PIR1bits.SSPIF = 0;
			
			/* Poll the current sensor (address shifted with r/w bit set) */
			tries = 0;
			SSPBUF = ((SensorAddr.asTable[type][leg] << 1) | 0b00000001);
			while (!PIR1bits.SSPIF)
			{
				if (++tries > TIME_OUT_CNT)
					break;
			}
			if (tries > TIME_OUT_CNT)
			{
				SNS_ResetI2C();
				continue;
			}
			PIR1bits.SSPIF = 0;	
			
			/* Make sure that we received an ACK from the sensor before reading
			 * a byte */
			if (!SSPCON2bits.ACKSTAT)
			{								
				/* Read a single byte */
				tries = 0;
				SSPCON2bits.RCEN = 1;
				while (!PIR1bits.SSPIF)
				{
					if(++tries > TIME_OUT_CNT)
						break;
				}
				if (tries > TIME_OUT_CNT)
				{
					SNS_ResetI2C();
					continue;
				}
				PIR1bits.SSPIF = 0;
				SensorReadings.asTable[type][leg] = SSPBUF;
		
				/* Send an NACK back to the slave (required by I2C before sending
				 * a stop/restart bit) */
				tries = 0;
				SSPCON2bits.ACKDT = 1;
				SSPCON2bits.ACKEN = 1;
				while (!PIR1bits.SSPIF)
				{
					if(++tries > TIME_OUT_CNT)
						break;
				}
				if (tries > TIME_OUT_CNT)
				{
					SNS_ResetI2C();
					continue;
				}
				PIR1bits.SSPIF = 0;
			}
			
			///TEST
			else
			{
				PORTCbits.RC1 ^=0;
			}		
			/* Generate a stop. */
			tries = 0;
			SSPCON2bits.PEN = 1;
			while (!PIR1bits.SSPIF)
			{
				if(++tries > TIME_OUT_CNT)
					break;
			}
			if (tries > TIME_OUT_CNT)
			{
				SNS_ResetI2C();
				continue;
			}
			PIR1bits.SSPIF = 0;
		}
	}
	
	/* Convert the latest readings to force values */
	Convert2Force();
	LowPassFilter();
		
	return;
}


/****************************** SNS_GetSingleRaw *****************************

Returns the raw sensor data from the given sensor type on the given leg.

Last Modified:
Barrett Heyneman	2006-8-9

*****************************************************************************/

unsigned char SNS_GetSingleRaw(Leg_t leg, SensorType_t type)
{
	return(SensorReadings.asTable[type][leg]);
}


/***************************** SNS_GetSingleForce ****************************

Returns the force data from the given sensor type on the given leg.

Last Modified:
Barrett Heyneman	2006-8-9

*****************************************************************************/

Number_t SNS_GetSingleForce(Leg_t leg, SensorType_t type)
{
	return(ForceValues.asTable[type][leg]);
}


/***************************** SNS_GetTypeRaw ********************************

Copies all raw sensor data for the given sensor type into the RawTypeValues_t
object pointed to by *dest.

Last Modified:
Barrett Heyneman	2006-8-9

*****************************************************************************/

void SNS_GetTypeRaw(SensorType_t type, SensorGroupRaw_t *dest)
{
	*dest = SensorReadings.asTypeArray[type];
}


/**************************** SNS_GetTypeForce *******************************

Copies all force data for the given sensor type into the ForceTypeValues_t
object pointed to by *dest.

LastModified:
Barrett Heyneman	2006-8-9

*****************************************************************************/

void SNS_GetTypeForce(SensorType_t type, SensorGroupNumber_t *dest)
{
	*dest = ForceValues.asTypeArray[type];
}


/****************************** SNS_GetAllRaw ********************************

Copies all the raw sensor readings into the RawValues_t object pointed to by
*dest.

Last Modified:
Barrett Heyneman 2006-8-9

*****************************************************************************/

void SNS_GetAllRaw(SensorRaw_t *dest)
{
	*dest = SensorReadings;
}


/****************************** SNS_GetAllRaw ********************************

Copies all the sensor forces into the ForceValues_t object pointed to by
*dest.

Last Modified:
Barrett Heyneman 2006-8-9

*****************************************************************************/

void SNS_GetAllForce(SensorNumber_t *dest)
{
	*dest = ForceValues;
}


/****************************** LowPassFilter ******************************

Low-pass filters the data

Last Modified:
Salomon Trujillo	2006-8-30

****************************************************************************/

static void LowPassFilter(void)
{

	//Multiply current value by 7/8 (using 1 tick = 1/256, 7/8 = 224 ticks)
	//and the previous value by 1/8 (32 ticks). The result will be shifted into the next byte
	//#define static_weight 32
	//#define incoming_weight 224

	#define static_weight 240
	#define incoming_weight 16

	static unsigned char i;
	static unsigned char j;
	
	NumberProduct_t val;

	for (i = 0; i < NUM_SENSOR_TYPES; i++)
	{
		for (j = 0; j < NUM_LEGS; j++)
		{
			val.raw = (((signed long)(incoming_weight)) * ((signed long)(ForceValues.asTable[i][j].val))) +
                      (((signed long)(static_weight)) * ((signed long)(filter[i][j])));
			
			filter[i][j] = val.formatted.result.raw;

			ForceValues.asTable[i][j] = val.formatted.result;
		}
	}
}


/****************************** Convert2Force *******************************

Converts the raw sensor readings (unsigned char) to force values (signed int)

Last Modified:
Barrett Heyneman	2006-8-4

****************************************************************************/

static void Convert2Force(void)
{
	static Number_t sensorDelta;
	static NumberProduct_t forceVal;
	static unsigned char i;
	static unsigned char j;
	
	/* Need to actually implement this at some point */
	for (i = 0; i < NUM_SENSOR_TYPES; i++)
	{
		for (j = 0; j < NUM_LEGS; j++)
		{
			sensorDelta.byParts.fraction = 0;
			sensorDelta.byParts.whole = SensorReadings.asTable[i][j] - SensorZeros.asTable[i][j];

			forceVal.raw = NumMul(sensorDelta, SensorSlopes.asTable[i][j]);
			
			/* See if the force value overflowed */
			if (!OverFlow(forceVal))
			{
				ForceValues.asTable[i][j] = forceVal.formatted.result;
			}
			
			/* If it did, set the min or max value allowable. */
			else if (forceVal.formatted.overflow > 0)
			{
				ForceValues.asTable[i][j].val = NUMBER_MAX;
			}
			else // if (forceVal.formatted.overflow < 0)
			{
				ForceValues.asTable[i][j].val = NUMBER_MIN;
			}
		}
	}
	
	return;
}

void SNS_ZeroSensors(void)
{
	static unsigned char i;
	static unsigned char j;

	for (i = 0; i < NUM_SENSOR_TYPES; i++)
	{
		for (j = 0; j < NUM_LEGS; j++)
		{
			 SensorZeros.asTable[i][j] = SensorReadings.asTable[i][j];
		}
	}
}
