/*****************************************************************************
*                              ControllerMath.h                              *
******************************************************************************

This module implements math functions for the number representation used in
all the math done by the controllers.  Currently that number representation is
as a signed int in base 1/256 (ie the decimal point lies between the two
bytes).

The C18 complier is little-endian, so the LSByte comes first.

*****************************************************************************/

#ifndef __CONTROLLER_MATH_H
#define __CONTROLLER_MATH_H

/*****************************************************************************
						PUBLIC TYPES, DEFINTIONS, MACROS
*****************************************************************************/

/* The number type we are using for all our computations.  The upper byte is
 * whole number value, while the lower byte is fracional (x/256).
 */

typedef union {
	unsigned int AsInt;
	struct {
		unsigned char LowerByte;
		unsigned char UpperByte;
	};
} uint16;

typedef union {
	signed int val;
	// *.raw added so all Number types can use IsZero() and SetZero()
	signed int raw;
	struct {
		unsigned char fraction;
		signed char whole;
	} byParts;
} Number_t;

/* In case Number_t.val has to be assigned directly, which hopefully won't
 * have to be done much, use this #define for casting (to keep abstraction).
 */
#define	Number_Type		signed int

/* Type used when multiplying two Number_t's.  Using the macros below, set the
 * 'raw' field to the result of Mul(n1,n2).  The fields 'formatted.underFlow'
 * and 'formatted.overFlow' allow you to check for under- (values with
 * magnitude smaller than 1/256) and over-flow (values outside 127+255/256 and
 * -128).  The sign of the overflow is the correct sign of the overall result.
 * The field 'formatted.product' is the result cast back to a Number_t.
 */
typedef union {
	signed long raw;
	struct {
		unsigned char underflow;
		Number_t result;
		signed char overflow;
	} formatted;
} NumberProduct_t;


/* Type used when adding/subtracting two Number_t's.  Using the macros below
 * set the 'raw' field to the result of Add(n1,n2) or Sub(n1,n2).  The field
 * 'formatted.overFlow' allows you to check for overflow.  Once again the
 * sign of the overflow byte is the sign of the overall result.  The field
 * 'formatted.sum' is the result cast back to a Number_t.
 */
typedef union {
	signed short long raw;
	struct {
		Number_t result;
		signed char overflow;
	} formatted;
} NumberSum_t;

/* Values which indicate no overflow in the 'formatted.overflow' field*/
#define POS_NO_OF			0x00
#define NEG_NO_OF			0xFF

/* Macros for adding, subtracting, and multiplication.  There is no division
 * macro because we're trying to keep things fast, and this implementation of
 * a number should eliminate the need for division. The macros are here so
 * we can build in the casting we need in one place and not have to remember
 * it everywhere we do math computations (ie multiplying two ints and storing
 * the result in a long requires casting one of the ints as a long first).
 */
#define NumAdd(n1, n2)		(signed short long)n1.val + n2.val
#define NumSub(n1, n2)		(signed short long)n1.val - n2.val
#define NumMul(n1, n2)		(signed long)n1.val * n2.val

/* Maximum and minimum values that a Number_t can hold */
#define NUMBER_MAX			(signed int)0x7FFF	// 127 + 255/256
#define NUMBER_MIN			(signed int)0x8000	// -128

/* Macro for checking for an overflow.  I put it here because it's a longish
 * line of code, and doing it as a macro helps with speed.  The macro works
 * with both NumberProduct_t and NumberSum_t, because the overflow byte is
 * accessed the same way for both.
 *
 * This works because if there was no overflow, all bits of the overflow byte
 * match the MSBit of the result; therefore, if you add that bit to the
 * overflow byte, you should get a (single byte) result of 0 iff there was no
 * overflow.
 */
#define OverFlow(value)	((unsigned char)value.formatted.overflow) + (value.formatted.result.byParts.whole & BIT7)

/* Macro to check if a value is zero (cleans up the code a little by not having
 * to directly access the *.raw fields of Number types.
 */
#define IsZero(num)		(num.raw == 0)

/* Macro to set a Number type to zero (again, cleaner by not having *.val or
 * *.raw accessed directly in code).
 */
#define SetZero(num)		(num.raw = 0)

/* Returns the absolute value of any of the number types (as the underlying
 * data type).
 */
#define Abs(num)		((num.raw < 0) ? -num.raw : num.raw)

/* The size (in bytes) of a Number_t (useful for reading in from gait files */
#define NUMBER_SIZE           sizeof(Number_t)


#endif /* __CONTROLLER_MATH_H */
