#include "servo.h"
#include "18fSensors.h"
#include "PhysicalConfig.h"
#include "Control.h"
#include "StanceController.h"
#include "Display.h"

#define true 1
#define false 0

#define NUM_STATES 4

//All positions are expressed as 2-byte numbers with the upper byte
//representing the 8-bit servo command and the lower byte is fractional
//data which is rounded off before transmission.

typedef enum {
	FLIGHT = 0,
	ATTACH = 1,
	STANCE = 2,
	DETACH = 3,
	DEFAULT_STATE = 255
} State_t;

const unsigned char NextToDetach[] = {1, 2, 3, 0};

static State_t LegState[NUM_LEGS];

//Positive foot: Curl to release
//Positive stroke: reach to head
//Positive wing: Pull off wall
#define STROKE_MAX 40
#define STROKE_MIN -55
#define ATTACH_STROKE_DIFF 52
#define LEG_SEPERATION 22

static ServoGroupNumber_t stroke_pos;
static ServoGroupNumber_t stroke_deltas;

//Function Prototypes
static void TransistionEvents(void);
static void StateControllers(void);

static unsigned char CtrlEnable;

void CTRL_ResetCtrl(void)
{
	ServoNumber_t InitValues;
	unsigned char i;

	for(i = 0; i < NUM_LEGS; i++)
		LegState[i] = STANCE;

	CTRL_SetupGesture(1, 0x37);
	LegState[1] = DETACH;

	InitValues.byName.FL_Foot.val = 0;
	InitValues.byName.FR_Foot.val = 0;
	InitValues.byName.BL_Foot.val = 0;
	InitValues.byName.BR_Foot.val = 0;
	InitValues.byName.FL_Stroke.byParts.whole = 8;
	InitValues.byName.FL_Stroke.byParts.fraction = 0;
	InitValues.byName.FR_Stroke.byParts.whole = -40;
	InitValues.byName.FR_Stroke.byParts.fraction = 0;
	InitValues.byName.BL_Stroke.byParts.whole = -24;
	InitValues.byName.BL_Stroke.byParts.fraction = 0;
	InitValues.byName.BR_Stroke.byParts.whole = -8;
	InitValues.byName.BR_Stroke.byParts.fraction = 0;
	InitValues.byName.FL_Wing.val = 0;
	InitValues.byName.FR_Wing.val = 0;
	InitValues.byName.BL_Wing.val = 0;
	InitValues.byName.BR_Wing.val= 0;

	SRV_SetAllAbs(&InitValues);
	SRV_GetGroupTargets(STROKE_SRV, &stroke_pos);

	CtrlEnable = false;
}

void CTRL_EnableCtrl(void)
{
	CtrlEnable = true;
}

void CTRL_DisableCtrl(void)
{
	CtrlEnable = false;
}

void CTRL_ControlLoop(void)
{
	if (CtrlEnable)
	{
		//Determine if the legs need to transistion between states
		StanceControl(&stroke_deltas);
		TransistionEvents();
		StateControllers();
	}
	else
	{
		StanceControl(&stroke_deltas);
		SRV_SetGroupAbs(STROKE_SRV, &stroke_deltas);
	}
}

static void TransistionEvents(void)
{
	unsigned char leg;
	
	for (leg = 0; leg < NUM_LEGS; leg++)
	{
		switch (LegState[leg])
		{
			case FLIGHT:
				if (SRV_GetSingleTarget(leg, STROKE_SRV).byParts.whole >= STROKE_MAX)
				{
					LegState[leg] = ATTACH;
					CTRL_SetupGesture(leg, 0x36);
				}
			break;
			case ATTACH:
				if (CTRL_GestureEOF(leg))
				{
					LegState[leg] = STANCE;
					stroke_pos.asValArray[leg].val = SRV_GetSingleTarget(leg, STROKE_SRV).val - stroke_deltas.asValArray[leg].val;

					LegState[NextToDetach[leg]] = DETACH;
					CTRL_SetupGesture(NextToDetach[leg], 0x37);
				}
			break;
			case DETACH:
				if (CTRL_GestureEOF(leg))
					LegState[leg] = FLIGHT;
			break;
		}
	}
}

static void StateControllers(void)
{
	const signed int flight_speed = 800;
	const signed int stance_speed = -100;

	Number_t servo_setting;

	unsigned char leg;


	for (leg = 0; leg < NUM_LEGS; leg++)
	{
		//Update stroke position
		switch (LegState[leg])
		{
			case FLIGHT:
				servo_setting.val = flight_speed;
				SRV_SetSingleRel(leg, STROKE_SRV, servo_setting);
			break;
			case STANCE:
				//servo_setting.val = stroke_pos.asValArray[leg].val + stroke_deltas.asValArray[leg].val;
				//SRV_SetSingleAbs(leg, STROKE_SRV, servo_setting);
				servo_setting.val = stance_speed;
				SRV_SetSingleRel(leg, STROKE_SRV, servo_setting);
			break;
			case ATTACH:
			case DETACH:
				CTRL_Gesture(leg);
			break;
		}
		stroke_pos.asValArray[leg].val += stance_speed;
	}
}

