#include "Control.h"
#include "FileReader.h"
#include "PhysicalConfig.h"
#include "servo.h"
#include "18fsensors.h"

static unsigned int gesture_byte_inx[NUM_LEGS];
static signed char gesture_delay_ticks[NUM_LEGS];
static unsigned char gesture_eof[NUM_LEGS];
static unsigned int gesture_phase[NUM_LEGS];
static Number_t pos[NUM_SERVO_TYPES][NUM_LEGS];
static Number_t vel[NUM_SERVO_TYPES][NUM_LEGS];
static Number_t acc[NUM_SERVO_TYPES][NUM_LEGS];
static Number_t force[NUM_SERVO_TYPES][NUM_LEGS];
static Number_t force_drv[NUM_SERVO_TYPES][NUM_LEGS];
static Number_t cmplnc_gain[NUM_SERVO_TYPES][NUM_LEGS];
static Number_t cmplnc_gain_drv[NUM_SERVO_TYPES][NUM_LEGS];

unsigned char CTRL_GestureEOF(unsigned char leg)
{
	return(gesture_eof[leg]);
}

void CTRL_SetupGesture(unsigned char leg, unsigned char gesture_index)
{
	static unsigned char i;

	//Reset all variables
	for(i = 0; i < NUM_SERVO_TYPES; i++)
	{
		pos[i][leg].val = SRV_GetSingleTarget(leg, i).val;
		vel[i][leg].val = 0;
		acc[i][leg].val = 0;
		force[i][leg] = SNS_GetSingleForce(leg, i);
		force[i][leg].val = 0;
		force_drv[i][leg].val = 0;
		cmplnc_gain[i][leg].val = 0;
		cmplnc_gain_drv[i][leg].val = 0;
	}
	gesture_phase[leg] = 0;

	//Get the location of the gesture in flash
	gesture_byte_inx[leg] = FR_FindEntry(gesture_index);

	//If the gesture can't be found, indicate so with an EOF flag.
	gesture_eof[leg] = (gesture_byte_inx[leg] == NULL_MEM_PTR);
	
	//Set the delay counter to one tick such that the initial decrement
	//upon entering CTRL_Gesture() sets the delay counter to zero
	gesture_delay_ticks[leg] = 1;
}

#define SERVO_POS	0x10
#define SERVO_VEL	0x20
#define SERVO_ACC	0x30
#define DSRD_FORCE	0x40
#define DSRD_FORCE_DRV	0x50
#define CMPLNCE_GN	0x60
#define CMPLNCE_GN_DRV	0x70
#define TIME_DELAY 	0x80
#define GESTURE_EOF 0x90

#define HI_NIBBLE 0xF0
#define LO_NIBBLE 0x0F

void CTRL_Gesture(unsigned char leg)
{
	static unsigned char srv_type;
	static unsigned char cmd;
	static unsigned int data;
	static NumberProduct_t prod;
	static Number_t delta;
	
	if (!gesture_eof[leg])
	{
		gesture_delay_ticks[leg]--;
		gesture_phase[leg]++;
	
		while(gesture_delay_ticks[leg] <= 0)
		{
			FR_ReadGestureData(&gesture_byte_inx[leg], &cmd, &data);
			switch (cmd & HI_NIBBLE)
			{
				case TIME_DELAY:
					gesture_delay_ticks[leg] = data;		
				break;
				case SERVO_POS:
					//Add some array checking here
					pos[cmd & LO_NIBBLE][leg].raw = data;
				break;
				case SERVO_VEL:
					//Add some array checking here
					vel[cmd & LO_NIBBLE][leg].raw = data;
				break;
				case SERVO_ACC:
					//Add some array checking here
					vel[cmd & LO_NIBBLE][leg].raw = data;
				break;
				case DSRD_FORCE:
					//Add some array checking here
					force[cmd & LO_NIBBLE][leg].raw = data;
				break;
				case DSRD_FORCE_DRV:
					//Add some array checking here
					force_drv[cmd & LO_NIBBLE][leg].raw = data;
				break;
				case CMPLNCE_GN:
					//Add some array checking here
					cmplnc_gain[cmd & LO_NIBBLE][leg].raw = data;
				break;
				case CMPLNCE_GN_DRV:
					//Add some array checking here
					cmplnc_gain_drv[cmd & LO_NIBBLE][leg].raw = data;
				break;
				case GESTURE_EOF:
					gesture_delay_ticks[leg] = 1;
					gesture_eof[leg] = 1;	
				break;
			}
		}
	}

	for(srv_type = 0; srv_type < NUM_SERVO_TYPES; srv_type++)
	{
		//Update terms based on their derivatives
		vel[srv_type][leg].val += acc[srv_type][leg].val;
		pos[srv_type][leg].val += vel[srv_type][leg].val;
		force[srv_type][leg].val += force_drv[srv_type][leg].val;
		cmplnc_gain[srv_type][leg].val += cmplnc_gain_drv[srv_type][leg].val;

		if (srv_type == STROKE_SRV)
		{
			delta.val = SNS_GetSingleForce(leg, STROKE_SNS).val - force[srv_type][leg].val;
			prod.raw = (signed long)cmplnc_gain[srv_type][leg].val * (signed long)delta.val;
			prod.formatted.result.val += pos[srv_type][leg].val;
			
			//Update servos
			SRV_SetSingleAbs(leg, srv_type, prod.formatted.result);
		}
		else
		{
			//Update servos
			SRV_SetSingleAbs(leg, srv_type, pos[srv_type][leg]);
		}
	}	
}
