/*****************************************************************************
 *                                   Keys.c                                  *
 *****************************************************************************
 
Key press routines for StickyBot Controller board (specifically for PIC18Fxxx)
Assumes that the keys are normally high (pulled-up in hardware).

Date:		Author:				Comment:
------------------------------------------------------------------------------
2003-10-28	Virgilio Mattoli	Original (v1.3 assembly code)
2006-5-8	Jonathan Karpick	Port to C
2006-7-3	Barrett Heyneman	General code cleanup

*****************************************************************************/

#include "Keys.h"
#include "18fUtil.h"
#include <delays.h>
#include <p18cxxx.h>


/*****************************************************************************
						PRIVATE TYPES, MACROS, DEFINITIONS
*****************************************************************************/

/* Hardware Settings */
#define KEY_PORT				PORTE		// All keys on Port E
#define KEY_PORT_TRIS			TRISE
#define KEY_PORT_TRIS_VAL		0b00000011	// Sets the two key pins to inputs
#define ADCON1_AD_MASK			0b00001111  // Mask to select the bits to control A/D status
#define ADCON1_MIN				0b00001010  // Minimum value for ADCON1

/* Timing */
#define DEBOUNCE_TIME			200			// Milliseconds

/* Key checking */
#define KEY_MASK				0b11111100	// Or with the port to see which keys are pressed


/*****************************************************************************
							PRIVATE VARIABLES
*****************************************************************************/
/* Current state of the keys */
static Byte_t curr_keys;


/*****************************************************************************
						PRIVATE FUNCTION PROTOTYPES
*****************************************************************************/


/*****************************************************************************
						PUBLIC FUNCTION DEFINITIONS	
*****************************************************************************/

/******************************** InitKeys ***********************************

Initializes the key pins as inputs

Last Modified
Barrett Heyneman	2006-7-3

*****************************************************************************/

void InitKeys(void)
{
	/* Use an '|=' because we are setting bits to 1's for input */
	KEY_PORT_TRIS |= KEY_PORT_TRIS_VAL;
	
	/* The ADCON1 value must be at least the value defined above (if it is
	 * higher than the PORTE pins are configured as digital inputs. */
	if ((ADCON1 & ADCON1_AD_MASK) < ADCON1_MIN)
	{
		ADCON1 = (ADCON1 & ~ADCON1_AD_MASK) | ADCON1_MIN;
	}
	
	return;
}


/******************************** CheckKeys **********************************

Returns a Byte_t with the bits set for all keys that are pressed.  If a key
was pressed, the code blocks for the debouncing period so that the keys aren't
polled too often (probably want to fix this somehow).

Last Modified:
Barrett Heyneman	2006-7-3

*****************************************************************************/

Byte_t CheckKeys(void)
{	
	curr_keys.byte = (KEY_PORT | KEY_MASK);
	if (curr_keys.byte)
		Milliseconds(DEBOUNCE_TIME);
		
	return(curr_keys);
}


/***************************** CheckSingleKey ********************************

Returns a 1 if the particular key is pressed, 0 if it isn't.

LastModified:
Barrett Heyneman	2006-7-3

*****************************************************************************/

unsigned char CheckSingleKey(Key_t key)
{
	curr_keys.byte = (KEY_PORT | KEY_MASK);
	
	switch (key)
	{
		case KEY_0:
			return((unsigned char)curr_keys.bit0);
		
		case KEY_1:
			return((unsigned char)curr_keys.bit1);
			
		default:
			return(0);
	}
}
