#ifndef _LINEARALGEBRA_HH_
#define _LINEARALGEBRA_HH_

#include <stdio.h>
#include <stdlib.h>

#define __TINY 1e-12
#define __MAX_MATRIX_SIZE 60
class LinearAlgebra {

  public:

  typedef enum { NO_ERROR=0, SING_MAT=1, MAT_TOO_BIG=2 } error_t;

  void printMatrix(double * mat, int numRows, int numCols)
  {
    int i, j;
    
    for(i = 0; i < numRows; i++)
    {
      for(j = 0; j < numCols; j++)
        printf("%8.3f  ", mat[i * numCols + j]);
      
      printf("\n");
    }
    printf("\n");
  
  }

  /* Example:
  If you want to evaluate the function M(A,B) = AB,
  then mat1 = A, mat2 = B, mat_out = M,
  If A is 3x5 and B is 5x2, then C is 3x2,
  and thus numRows = 3, numCommon = 5 and numCols = 2*/
  int multiplyMatrix(double * mat1, double * mat2, double * mat_out,
                     int numRows, int numCommon, int numCols)
  {
    int i, j, k;
    
    for(i = 0; i < numRows; i++)
      for(j = 0; j < numCols; j++)
      {
        mat_out[i * numCols + j] = 0.0;
      
        for(k = 0; k < numCommon; k++)
          mat_out[i * numCols + j] += mat1[i * numCommon + k] * mat2[k * numCols + j];
	
      }
    
    return NO_ERROR;
  }

  int detMatrix(double * mat, double * det, int size)
  {
    int k;
    int _row[25], _col[25];

    if (size > __MAX_MATRIX_SIZE)
      return MAT_TOO_BIG;

    for (k = 0; k < size; k++)
    {
      _row[k] = k;
      _col[k] = k;
    }
    
    *det = det_iter(mat, _row, _col, size, size);
    
    return NO_ERROR;
  }
  
  int inverseMatrix(double * mat, double * inv_mat, int size)
  {
    double det;
    int i, j, k;
    int _row[25], _col[25];

    if (size > __MAX_MATRIX_SIZE)
      return MAT_TOO_BIG;

    for (k = 0; k < size; k++)
    {
      _row[k] = k;
      _col[k] = k;
    }
    det = det_iter(mat, _row, _col, size, size);

    if ((det < -__TINY) || (__TINY < det))
    {
      for (i = 0; i < size; i++)
        for (j = 0; j < size; j++)
        {
          for (k = 0; k <= (i - 1); k++)
            _row[k] = k;
          for (k = i+1; k < size; k++)
            _row[k-1] = k;
	
          for (k = 0; k <= (j - 1); k++)
            _col[k] = k;
          for (k = j+1; k < size; k++)
            _col[k-1] = k;

          inv_mat[j * size + i] = ( (((i+j) % 2) == 0) ? 1 : -1) *
                                  det_iter(mat, (int *) _row, (int *) _col, size - 1, size)
		      	          / det;
         }
       return NO_ERROR;
     }
     else
     {
       return SING_MAT;
     }
  }

  private:

  double det_iter(double * mat, int * row, int * col, int size, int mSize)
  {
    double rtn;
    int _row[25], _col[25];
    int i, k;

    if (size == 1)
      rtn = mat[col[0] + mSize * row[0]];
    else
    {
      rtn = 0.0;
    
      for (i = 0; i < size; i++)
      {
        for (k = 1; k < size; k++)
          _row[k-1] = row[k];

        for (k = 0; k <= (i - 1); k++)
          _col[k] = col[k];
        for (k = i+1; k < size; k++)
          _col[k-1] = col[k];

        rtn += mat[row[0] * mSize + col[i]] *
               ( (((i) % 2) == 0) ? 1 : -1) *
	       det_iter(mat, _row, _col, size - 1, mSize);
      }
    }  
    return rtn;
  }

};
#endif

/* Test main
int main(void)
{
  const int size = 5;
  int i, j;
  double mat[size * size];
  double inv_mat[size * size];

  LinearAlgebra * LnAlg = new LinearAlgebra;

  for(i = 0; i < size; i++)
    for(j = 0; j < size; j++)
      mat[i * size + j] = 25.0 * (double)rand() / RAND_MAX;

  LnAlg->printMatrix(mat, size, size);
  printf("\n");
  
  LnAlg->inverseMatrix(mat, inv_mat, size);
  LnAlg->printMatrix(inv_mat, size, size);
  printf("\n\n");

  const int row = 3;
  const int common = 5;
  const int col = 2;
  double mat1[row * common];
  double mat2[common * col];
  double mat_out[row * col];
  
  for(i = 0; i < row; i++)
    for(j = 0; j < common; j++)
      mat1[i * size + j] = 25.0 * (double)rand() / RAND_MAX;

  for(i = 0; i < common; i++)
    for(j = 0; j < col; j++)
      mat2[i * size + j] = 25.0 * (double)rand() / RAND_MAX;
  
  LnAlg->printMatrix(mat1, row, common);
  printf("\n");
  LnAlg->printMatrix(mat2, common, col);
  printf("\n");
  LnAlg->multiplyMatrix(mat1, mat2, mat_out, row, common, col);
  LnAlg->printMatrix(mat_out, row, col);

  return 0;
}
*/
