/*****************************************************************************
*                           StanceController.c                               *
******************************************************************************

Date:		Author:				Comment:
------------------------------------------------------------------------------
2006-8-1	Barrett Heyneman	Original Version
2006-8-10	Barrett Heyneman	Reworked to take advantage of the new Number,
								Sensor, and Servo data types.

*****************************************************************************/

#include "StanceController.h"
#include "18fSensors.h"
#include "18fUtil.h"
#include "18f4520_flash.h"
#include "MatrixMath.h"
#include "ControllerMath.h"
#include "PhysicalConfig.h"
#include "FileReader.h"

/*****************************************************************************
						PRIVATE TYPES, MACROS, DEFINITIONS
*****************************************************************************/


/*****************************************************************************
							PRIVATE VARIABLES
*****************************************************************************/

/* The grasp matrix and its inverse */
static Number_t Grasp[4][4];
static Number_t InvGrasp[4][4];

/* The grasp-space gains */
static Number_t Gains[4];

/* The desired forces */
static Number_t FDes[4];

/* The feed forward velocities */
static Number_t VelFF[4];


/*****************************************************************************
						PRIVATE FUNCTION DEFINITIONS
*****************************************************************************/


/*****************************************************************************
						PUBLIC FUNCTION DEFINITIONS
*****************************************************************************/


/**************************** InitStanceControl ******************************

Reads in the grasp matrix and its inverse, the grasp space gains, desired
grasp space forces, and the initial grasp space feed-forward velocity from
flash and loads them in local variables.

Last Modified:
Barrett Heyneman	2006-8-2

*****************************************************************************/
 
void InitStanceControl(void)
{
	near unsigned short long read_loc = TOC_START;
	
	read_loc = (near unsigned short long)FR_FindEntry(STANCE_CONTROLLER);
	if (read_loc == NO_ENTRY_FOUND)
		/* Change to an error code */
		return;
		
	/* Grab all the data */
	
	/* First 16 bytes of data are the Grasp Matrix */
	ReadFlash(read_loc, &Grasp, 16*NUMBER_SIZE);
	read_loc += 16*NUMBER_SIZE;
	
	/* Next 16 are the inverse grasp matrix */
	ReadFlash(read_loc, &InvGrasp, 16*NUMBER_SIZE);
	read_loc += 16*NUMBER_SIZE;
	
	/* Next 4 are the gains */
	ReadFlash(read_loc, &Gains, 4*NUMBER_SIZE);
	read_loc += 4*NUMBER_SIZE;
	
	/* Next 4 are the desired forces */
	ReadFlash(read_loc, &FDes, 4*NUMBER_SIZE);
	read_loc += 4*NUMBER_SIZE;
	
	/* Next 4 are the feed forward velocities */
	ReadFlash(read_loc, &VelFF, 4*NUMBER_SIZE);
	
	return;
}


/****************************** StanceControl ********************************

Applies the stance control law to all the legs and puts the computed servo
deltas into memory starting at address 'deltas' (just the stroke servos)

At last check, this controller code took 648 us to execute.

Last Modified:
Barrett Heyneman	2006-8-29

*****************************************************************************/

void StanceControl(ServoGroupNumber_t *stroke_deltas)
{
	static SensorGroupNumber_t forces;
	static Number_t temp[NUM_LEGS];
	static Number_t temp2[NUM_LEGS];
	
	/* Read in the stroke motor forces. */
	SNS_GetTypeForce(STROKE_SNS, &forces);
	
	/* Find the difference between actual and desired grasp space forces */
	VectorSub(forces.asValArray, FDes, temp);

	/* Convert to grasp space by left multiplying by the grasp matrix */
	MatrixVector(Grasp, temp, temp2);
	
	///TEST
	/* Kill the first FDes term (total force) so that we don't have to estimate the weight yet */
	SetZero(temp2[0]);
	
	/* Multiply that by the diagonal gain matrix (represented as a vector) */
	VectorMul(temp2, Gains, temp);
	
	/* Add in the feed forward grasp space velocity */
	VectorAdd(temp, VelFF, temp);
	
	/* Multiply by the inverse grasp matrix to get deltas */
	MatrixVector(InvGrasp, temp, stroke_deltas);
	
	return;
}

void SetForceDesired(SensorGroupNumber_t *force_desired)
{
	static unsigned i;

	for(i = 0; i < NUM_LEGS; i++)
		FDes[i] = force_desired->asValArray[i];
}

void GetForceDesired(SensorGroupNumber_t *force_desired)
{
	static unsigned i;

	for(i = 0; i < NUM_LEGS; i++)
		force_desired->asValArray[i] = FDes[i];
}

/*
void TestInit()
{
	Grasp[0][0].val = 0x0200;
	Grasp[0][1].val = 0x0000;
	Grasp[0][2].val = 0x0000;
	Grasp[0][3].val = 0x0000;
	Grasp[1][0].val = 0x0100;
	Grasp[1][1].val = 0x0100;
	Grasp[1][2].val = 0x0000;
	Grasp[1][3].val = 0x0000;
	Grasp[2][0].val = 0x0000;
	Grasp[2][1].val = 0x0000;
	Grasp[2][2].val = 0x0200;
	Grasp[2][3].val = 0x0000;
	Grasp[3][0].val = 0x0000;
	Grasp[3][1].val = 0x0000;
	Grasp[3][2].val = 0x0100;
	Grasp[3][3].val = 0x0100;
	
	InvGrasp[0][0].val = 0xFF00;
	InvGrasp[0][1].val = 0x0000;
	InvGrasp[0][2].val = 0x0000;
	InvGrasp[0][3].val = 0x0200;
	InvGrasp[1][0].val = 0x0000;
	InvGrasp[1][1].val = 0x0100;
	InvGrasp[1][2].val = 0x0000;
	InvGrasp[1][3].val = 0x0000;
	InvGrasp[2][0].val = 0x0000;
	InvGrasp[2][1].val = 0x0000;
	InvGrasp[2][2].val = 0xFF00;
	InvGrasp[2][3].val = 0x0000;
	InvGrasp[3][0].val = 0x0200;
	InvGrasp[3][1].val = 0x0000;
	InvGrasp[3][2].val = 0x0000;
	InvGrasp[3][3].val = 0x0100;
	
	Gains[0].val = 0x0080;
	Gains[1].val = 0x0300;
	Gains[2].val = 0x0100;
	Gains[3].val = 0xFF00;
	
	FDes[0].val = 0x0500;
	FDes[1].val = 0x0000;
	FDes[2].val = 0x0000;
	FDes[3].val = 0x0000;
	
	VelFF[0].val = 0x0200;
	VelFF[1].val = 0x0000;
	VelFF[2].val = 0xFF00;
	VelFF[3].val = 0x0000;
}
*/
